import codecs
import copy
import json
import os

from tgalice.storage.database_utils import fix_bson_keys
from tgalice.utils import database_utils


class BaseStorage:
    def __init__(self):
        self.dict = {}

    def get(self, key):
        return copy.deepcopy(self.dict.get(key, {}))

    def set(self, key, value):
        self.dict[key] = copy.deepcopy(value)


class FileBasedStorage(BaseStorage):
    def __init__(self, path='session_storage', multifile=True):
        super(FileBasedStorage, self).__init__()
        self.path = path
        self.multifile = multifile
        if not os.path.exists(path):
            if self.multifile:
                os.mkdir(path)
            else:
                self.dump_dict(path, {})

    def dump_dict(self, filename, data):
        with open(filename, 'w', encoding='utf-8') as f:
            json.dump(data, f, indent=2, ensure_ascii=False)

    def load_dict(self, filename):
        if not os.path.exists(filename):
            return {}
        with open(filename, 'r', encoding='utf-8') as f:
            result = json.load(f)
        return result

    def get(self, key):
        if self.multifile:
            return self.load_dict(os.path.join(self.path, key))
        else:
            return self.load_dict(self.path).get(key, {})

    def set(self, key, value):
        if self.multifile:
            self.dump_dict(os.path.join(self.path, key), value)
        else:
            # todo: enable some concurrency guarantees
            data = self.load_dict(self.path)
            data[key] = value
            self.dump_dict(self.path, data)


class MongoBasedStorage(BaseStorage):
    KEY_NAME = 'key'
    VALUE_NAME = 'value'

    def __init__(self, database=None, collection_name='sessions', collection=None):
        assert database or collection
        super(MongoBasedStorage, self).__init__()
        # we assume that the database has PyMongo interface
        self._collection = collection or database.get_collection(collection_name)
        database_utils.ensure_mongo_index(index_name=self.KEY_NAME, collection=self._collection)

    def get(self, key):
        result = self._collection.find_one({self.KEY_NAME: key})
        if result is None:
            return {}
        return result.get(self.VALUE_NAME, {})

    def set(self, key, value):
        value = fix_bson_keys(value)
        self._collection.update_one(
            {self.KEY_NAME: key},
            {'$set': {self.VALUE_NAME: value}},
            upsert=True
        )


class S3BasedStorage(BaseStorage):
    """ This wrapper is intended to work with a boto3 client - e.g. in Yandex.Cloud Object Storage. """
    def __init__(self, s3_client, bucket_name, prefix=''):
        super(BaseStorage, self).__init__()
        self.s3_client = s3_client
        self.bucket_name = bucket_name
        self.prefix = prefix

    def modify_key(self, key):
        return self.prefix + key

    def get(self, key):
        try:
            result = self.s3_client.get_object(Bucket=self.bucket_name, Key=self.modify_key(key))
            body = result['Body']
            reader = codecs.getreader("utf-8")
            return json.load(reader(body))
        except Exception as e:
            if hasattr(e, 'response') and e.response.get('Error', {}).get('Code') == 'NoSuchKey':
                return {}
            else:
                raise e

    def set(self, key, value):
        self.s3_client.put_object(Bucket=self.bucket_name, Key=self.modify_key(key), Body=json.dumps(value))



import re
import random

_RANDOM: bool = True


def random_on():
    global _RANDOM
    _RANDOM = True


def random_off():
    global _RANDOM
    _RANDOM = False


def sample(pattern, rnd=None, sep='|', drop_proba=0.5):
    """ Sample various texts from patterns such as '{Okay|well}{ dude}*, when should we {begin|start}?'
    Group of options are enclosed by '{}' brackets and separated by `sep`. Nested groups are not supported.
    The groups marked by '*' star are omitted with probability `drop_proba`.
    If `rnd` is False, the first option is always returned, which may simplify testing or debugging.
    """
    if rnd is None:
        rnd = _RANDOM

    def f(x):
        options = x.group(1).split(sep)
        if not options:
            return ''
        if not rnd:
            return options[0]
        if drop_proba == 1:
            return ''
        elif drop_proba == 0:
            pass
        elif x.group().endswith('*') and random.random() > drop_proba:
            return ''
        return random.choice(options)

    return re.sub('{([^}]*?)}\\*?', f, pattern)



import tgalice

from tgalice.testing.testing_utils import make_context


DEFAULT_MESSAGE = 'this is the default message'


def test_faq():
    dm = tgalice.dialog_manager.FAQDialogManager(
        'tests/test_managers/faq.yaml',
        matcher='cosine',
        default_message=DEFAULT_MESSAGE
    )
    r1 = dm.respond(make_context(new_session=True))
    assert r1.text == DEFAULT_MESSAGE
    first_responses = {dm.respond(make_context(text='hi there', prev_response=r1)).text for i in range(30)}
    assert first_responses == {'Hello!', 'Nice to meet you!', 'Have a good time, human.'}

    r2 = dm.respond(make_context(text='hi there', prev_response=r1))
    assert set(r2.suggests) == {'How are you?', 'What can you do?'}

    r3 = dm.respond(make_context(text='how are you', prev_response=r2))
    assert r3.text == "I'm fine, thanks"

    r4 = dm.respond(make_context(text='What can you do?', prev_response=r3))
    assert r4.text == DEFAULT_MESSAGE



from tgalice.utils.collections import make_unique, sample_at_most

import random


def test_make_unique():
    random.seed(42)
    for n in range(100):
        x = [random.randrange(10) for i in range(n)]
        y = make_unique(x)
        seen = set()
        for element in x:
            if element not in seen:
                assert y[len(seen)] == element
                seen.add(element)


def test_sample_at_most():
    random.seed(42)
    for n in range(100):
        for m in range(100):
            x = [random.randrange(10) for i in range(n)]
            y = sample_at_most(x, m)
            assert len(set(y)) == len(y)
            assert len(y) == min(m, len(set(x)))
            assert not set(y).difference(set(x))



import pytest
from tgalice.cascade import Cascade, DialogTurn
from tgalice.dialog import Context
from tgalice.testing.testing_utils import make_context

csc = Cascade()


@csc.add_handler(priority=0)
def t0(turn: DialogTurn):
    turn.response_text = '0'


@csc.add_handler(priority=1, intents=['i1'])
def t1(turn: DialogTurn):
    turn.response_text = '1'


@csc.add_handler(priority=1, intents=['i2'])
def t2(turn: DialogTurn):
    turn.response_text = '2'


@csc.add_handler(priority=2, intents=['i3'])
def t3(turn: DialogTurn):
    turn.response_text = '2'


@csc.add_handler(priority=3, stages=['s1'])
def t4(turn: DialogTurn):
    turn.response_text = 'stage 1'


@csc.postprocessor
def ask_for_tea(turn: DialogTurn):
    turn.response_text += '\nDo you want some tea?'


@pytest.mark.parametrize('intents,result', [
    ({}, 't0'),
    ({'i1': 1.0, 'i2': 0.5}, 't1'),
    ({'i2': 0.5}, 't2'),
    ({'i1': 1.0, 'i2': 0.5, 'i3': 0.1}, 't3'),
])
def test_ranking(intents, result):
    turn = DialogTurn(make_context(text='kek'), text='kek', intents=intents)
    assert csc(turn) == result


def test_ranking_stage():
    ctx = Context(message_text='kek', user_object={'stage': 's1'}, metadata=None)
    turn = DialogTurn(ctx, text='kek', intents={'i3': 1})
    assert turn.old_user_object == {'stage': 's1'}
    assert turn.stage == 's1'
    assert csc(turn) == 't4'


def test_postprocess():
    turn = DialogTurn(make_context(text='kek'), text='kek')
    turn.response_text = 'The weather is cool.'
    # without agenda, no postprocessors are called
    turn.release_control()
    csc.postprocess(turn)
    assert turn.response_text.endswith('cool.')
    # without control, no postprocessors are called
    turn.take_control()
    turn.add_agenda('ask_for_tea')
    csc.postprocess(turn)
    assert turn.response_text.endswith('cool.')
    # with control and agenda, postprocessors are called
    turn.release_control()
    csc.postprocess(turn)
    assert turn.response_text.endswith('tea?')
    # after postprocessing, agenda goes away
    assert not turn.agenda



import pytest

from tgalice.dialog_connector import DialogConnector, SOURCES
from tgalice.dialog_manager import BaseDialogManager, Response


class Repeater(BaseDialogManager):
    def respond(self, ctx):
        return Response(text=ctx.message_text)


@pytest.fixture
def alice_message():
    return {
        'meta': {
            'client_id': 'ru.yandex.searchplugin/7.16 (none none; android 4.4.2)',
            'interfaces': {
                'account_linking': {},
                'payments': {},
                'screen': {}
            },
            'locale': 'ru-RU',
            'timezone': 'UTC'
        },
        'request': {
            'command': 'привет',
            'nlu': {
                'entities': [],
                'tokens': ['привет']
            },
            'original_utterance': 'передай тестовому навыку привет',
            'type': 'SimpleUtterance'
        },
        'session': {
            'message_id': 0,
            'new': True,
            'session_id': 'session-1',
            'skill_id': 'skill-1',
            'user_id': 'user-id-1'
        },
        'version': '1.0'
    }


def test_alice_input(alice_message):
    connector = DialogConnector(Repeater())
    ctx = connector.make_context(message=alice_message, source=SOURCES.ALICE)
    assert ctx.user_id == 'alice__user-id-1'
    assert ctx.message_text == 'привет'
    assert ctx.metadata['new_session']


def test_alice_response(alice_message):
    connector = DialogConnector(Repeater())
    response = Response(text='привет', suggests=['тебе тоже'])
    output = connector.standardize_output(source=SOURCES.ALICE, original_message=alice_message, response=response)
    assert not output['response'].get('end_session')
    assert output['response']['text'] == 'привет'
    assert 'voice' not in output['response'] or output['response']['voice'] == 'привет'
    assert output['response']['buttons'] == [{'title': 'тебе тоже', 'hide': True}]


def test_user_objects():
    connector = DialogConnector(Repeater())
    assert connector.get_user_object('user_1') == {}
    connector.set_user_object('user_1', {'name': 'alex'})
    assert connector.get_user_object('user_2') == {}
    assert connector.get_user_object('user_1') == {'name': 'alex'}


def test_save_user_object_by_yandex(alice_message):
    connector = DialogConnector(Repeater(), alice_native_state=True)
    uo = {
        'session': {'count': 1},
        'user': {'name': 'Alex'}
    }
    response = Response(text='привет', user_object=uo)
    output = connector.standardize_output(source=SOURCES.ALICE, original_message=alice_message, response=response)
    assert output['session_state'] == {'count': 1}
    assert output['user_state_update'] == {'name': 'Alex'}


def test_save_user_object_by_yandex_to_session(alice_message):
    connector = DialogConnector(Repeater(), alice_native_state='session')
    uo = {'count': 1, 'name': 'Alex'}
    response = Response(text='привет', user_object=uo)
    output = connector.standardize_output(source=SOURCES.ALICE, original_message=alice_message, response=response)
    assert output['session_state'] == {'count': 1, 'name': 'Alex'}


def test_save_user_object_by_yandex_to_user(alice_message):
    connector = DialogConnector(Repeater(), alice_native_state='user')
    uo = {'count': 1, 'name': 'Alex'}
    response = Response(text='привет', user_object=uo)
    output = connector.standardize_output(source=SOURCES.ALICE, original_message=alice_message, response=response)
    assert output['user_state_update'] == {'count': 1, 'name': 'Alex'}


@pytest.mark.parametrize('ans,uo', [
    (True, {
        'session': {'value': 10},
        'user': {'value': 42}
    }),
    ('session', {'value': 10}),
    ('user', {'value': 42}),
    (False, {}),
])
def test_load_user_object_by_yandex(ans, uo, alice_message):
    connector = DialogConnector(Repeater(), alice_native_state=ans)
    alice_message['state'] = {
        'session': {
            'value': 10
        },
        'user': {
            'value': 42,
        }
    }
    ctx = connector.make_context(message=alice_message, source=SOURCES.ALICE)
    assert ctx.user_object == uo



import random
from typing import Mapping

from tgalice.dialog import Response, names


class Phrase:
    """ This is a template for Response
    It has some methods to be parsed from configs.
    It generates a Response possibly by sampling options.
    """
    def __init__(self, name, text, exit=False, suggests=None):
        self.name = name
        if isinstance(text, list):
            self.texts = text
        else:
            self.texts = [text]
        self.exit = exit
        self.suggests = suggests or []
        # todo: read everything that can be found in response config

    def render(self, seed=None, additional_suggests=None):
        if seed is not None:
            random.seed(seed)
        text = random.choice(self.texts)
        resp = Response(text, suggests=self.suggests)
        resp.set_text(text)
        if self.exit:
            resp.commands.append(names.COMMANDS.EXIT)
        if additional_suggests:
            resp.suggests.extend(additional_suggests)
        return resp

    @classmethod
    def from_object(cls, obj):
        if isinstance(obj, str):
            return cls(name='unnamed', text=obj)
        elif isinstance(obj, Mapping):
            return cls(**obj)
        elif isinstance(obj, cls):
            return obj
        else:
            raise ValueError('Type {} cannot be converted to Phrase'.format(type(obj)))



import pymorphy2
import re

from functools import lru_cache

PYMORPHY = pymorphy2.MorphAnalyzer()


@lru_cache(maxsize=16384)
def word2lemma(word):
    hypotheses = PYMORPHY.parse(word)
    if len(hypotheses) == 0:
        return word
    return hypotheses[0].normal_form


def fast_normalize(text, lemmatize=False):
    text = re.sub('[^a-zа-яё0-9]+', ' ', text.lower())
    # we consider '-' as a delimiter, because it is often missing in results of ASR
    text = re.sub('\\s+', ' ', text).strip()
    if lemmatize:
        text = ' '.join([word2lemma(w) for w in text.split()])
    text = re.sub('ё', 'е', text)
    return text


def like_help(text):
    text = fast_normalize(text)
    return bool(re.match('^(алиса |яндекс )?(помощь|что ты (умеешь|можешь))$', text))


def like_exit(text):
    text = fast_normalize(text)
    return bool(re.match('^(алиса |яндекс )?(выход|хватит( болтать| играть)?|выйти|закончить)$', text))


def like_yes(text):
    text = fast_normalize(text)
    return bool(re.match('^(да|ага|окей|ок|конечно|yes|yep|хорошо|ладно)$', text))


def like_no(text):
    text = fast_normalize(text)
    return bool(re.match('^(нет|не|no|nope)$', text))



import attr
import logging

from tgalice.dialog import Context, Response
from tgalice.utils.content_manager import YandexImageAPI
from tgalice.nlg.controls import Gallery as VisualGallery, BigImage
from typing import Dict, List, Optional, Union, Tuple

logger = logging.getLogger(__name__)


FORM_TYPE = Dict[str, Dict]


@attr.s
class DialogTurn:
    """ DialogTurn is a single-variable wrapper for both context and response """
    ctx: Context = attr.ib()
    # These properties are extracted from context:
    text: str = attr.ib()
    intents: Dict = attr.ib(factory=dict)
    forms: Dict[str, Dict] = attr.ib(factory=dict)
    # These properties will be put into response:
    response_text: str = attr.ib(default='')
    response: Optional[Response] = attr.ib(default=None)
    user_object = attr.ib(factory=dict)
    suggests: List = attr.ib(factory=list)
    commands: List = attr.ib(factory=list)
    card: Optional[Union[VisualGallery, BigImage]] = attr.ib(default=None)
    image_url: Optional[str] = attr.ib(default=None)
    # These properties are helpers
    image_manager: Optional[YandexImageAPI] = attr.ib(default=None)
    can_change_topic: bool = attr.ib(default=False)

    _STAGE = 'stage'
    _AGENDA = 'agenda'
    _AGENDA_FORMS = 'agenda_forms'

    @property
    def old_user_object(self):
        return self.ctx.user_object

    @property
    def is_complete(self):
        return bool(self.response_text) or bool(self.response)

    @property
    def stage(self) -> Optional[str]:
        return self.old_user_object.get(self._STAGE)

    @property
    def next_stage(self) -> Optional[str]:
        return self.user_object.get(self._STAGE)

    @stage.setter
    def stage(self, value):
        self.user_object[self._STAGE] = value

    def release_control(self):
        self.can_change_topic = True

    def take_control(self):
        self.can_change_topic = False

    @property
    def can_take_control(self) -> bool:
        return self.can_change_topic and not self.next_stage

    @property
    def agenda(self) -> List[str]:
        """ The stack of postprocessors waiting to be triggered """
        return self.user_object.get(self._AGENDA, [])

    @property
    def agenda_forms(self) -> FORM_TYPE:
        """ The forms to be fed into postprocessors """
        return self.user_object.get(self._AGENDA_FORMS, {})

    def add_agenda(self, postprocessor_name: str, form: Dict = None) -> bool:
        if not self.agenda:
            self.user_object[self._AGENDA] = []
        if postprocessor_name not in self.agenda:
            # avoiding duplicate agenda and potential cycles
            self.agenda.append(postprocessor_name)
            if form:
                if not self.agenda_forms:
                    self.user_object[self._AGENDA_FORMS] = {}
                self.agenda_forms[postprocessor_name] = form
            return True
        return False

    def pop_agenda(self) -> Tuple[Optional[str], Optional[FORM_TYPE]]:
        if not self.agenda:
            self.clear_agenda()
            return None, None
        key = self.agenda.pop()
        form = self.agenda_forms.get(key)
        if form:
            del self.agenda_forms[key]
        return key, form

    def clear_agenda(self):
        self.user_object[self._AGENDA] = []
        self.user_object[self._AGENDA_FORMS] = {}

    def add_space(self):
        if not self.response_text:
            self.response_text = ''
        else:
            self.response_text += '\n'

    def make_response(self) -> Optional[Response]:
        if self.response:
            return self.response
        if self.response_text:
            r = Response(
                text=None,
                user_object=self.user_object,
                rich_text=self.response_text,
                suggests=self.suggests,
                commands=self.commands,
                image_url=self.image_url,
            )
            if isinstance(self.card, BigImage):
                r.image = self.card
            elif isinstance(self.card, VisualGallery):
                r.gallery = self.card
            return r



import telebot

import tgalice.utils

from typing import Optional

from tgalice.storage.session_storage import BaseStorage
from tgalice.dialog_manager.base import Response, Context
from tgalice.dialog.names import COMMANDS, SOURCES
from tgalice.interfaces.yandex import YandexResponse
from tgalice.utils.content_manager import YandexImageAPI


class DialogConnector:
    """ This class provides unified interface for both Telegram and Alice applications """
    def __init__(
            self,
            dialog_manager,
            storage=None,
            log_storage=None,
            default_source=SOURCES.TELEGRAM,
            tg_suggests_cols=1,
            alice_native_state=False,
            image_manager: Optional[YandexImageAPI] = None,
    ):
        """
        paramaters:
        - alice_native_state: bool or 'user' or 'state'
        """
        self.dialog_manager = dialog_manager
        self.default_source = default_source
        self.storage = storage or BaseStorage()
        self.log_storage = log_storage  # noqa
        self.tg_suggests_cols = tg_suggests_cols
        self.alice_native_state = alice_native_state
        self.image_manager = image_manager

    def respond(self, message, source=None):
        # todo: support different triggers - not only messages, but calendar events as well
        context = self.make_context(message=message, source=source)
        if self.log_storage is not None:
            self.log_storage.log_context(context)

        response = self.dialog_manager.respond(context)
        if response.updated_user_object is not None and response.updated_user_object != context.user_object:
            if source == SOURCES.ALICE and self.alice_native_state:
                pass  # user object is added right to the response
            else:
                self.set_user_object(context.user_id, response.updated_user_object)

        result = self.standardize_output(source, message, response)
        if self.log_storage is not None:
            self.log_storage.log_response(data=result, context=context, source=context.source, response=response)
        return result

    def make_context(self, message, source=None):
        if source is None:
            source = self.default_source
        context = Context.from_raw(source=source, message=message)
        if source == SOURCES.ALICE and self.alice_native_state:
            if self.alice_native_state == 'session':
                user_object = message.get('state', {}).get('session')
            elif self.alice_native_state == 'user':
                user_object = message.get('state', {}).get('user')
            else:
                user_object = message.get('state')
        else:
            user_object = self.get_user_object(context.user_id)
        context.add_user_object(user_object)
        return context

    def get_user_object(self, user_id):
        if self.storage is None:
            return {}
        return self.storage.get(user_id)

    def set_user_object(self, user_id, user_object):
        if self.storage is None:
            raise NotImplementedError()
        self.storage.set(user_id, user_object)

    def standardize_output(self, source, original_message, response: Response):
        has_exit_command = False
        if response.commands:
            for command in response.commands:
                if command == COMMANDS.EXIT:
                    has_exit_command = True
                else:
                    raise NotImplementedError('Command "{}" is not implemented'.format(command))
        if source == SOURCES.TELEGRAM:
            if response.raw_response is not None:
                return response.raw_response
            result = {
                'text': response.text
            }
            if response.links is not None:
                result['parse_mode'] = 'html'
                for link in response.links:
                    result['text'] += '\n<a href="{}">{}</a>'.format(link['url'], link['title'])
            if response.suggests:
                # todo: do smarter row width calculation
                row_width = min(self.tg_suggests_cols, len(response.suggests))
                result['reply_markup'] = telebot.types.ReplyKeyboardMarkup(row_width=row_width)
                result['reply_markup'].add(*[telebot.types.KeyboardButton(t) for t in response.suggests])
            else:
                result['reply_markup'] = telebot.types.ReplyKeyboardRemove(selective=False)
            if response.image_url:
                if 'multimedia' not in result:
                    result['multimedia'] = []
                media_type = 'document' if response.image_url.endswith('.gif') else 'photo'
                result['multimedia'].append({'type': media_type, 'content': response.image_url})
            if response.sound_url:
                if 'multimedia' not in result:
                    result['multimedia'] = []
                result['multimedia'].append({'type': 'audio', 'content': response.sound_url})
            result['disable_web_page_preview'] = True
            return result
        elif source == SOURCES.ALICE:
            result = {
                "version": original_message['version'],
                "response": {
                    "end_session": has_exit_command,
                    "text": response.text
                }
            }
            if self.alice_native_state and response.updated_user_object:
                if self.alice_native_state == 'session':
                    result['session_state'] = response.updated_user_object
                elif self.alice_native_state == 'user':
                    result['user_state_update'] = response.updated_user_object
                else:
                    if 'session' in response.updated_user_object:
                        result['session_state'] = response.updated_user_object['session']
                    if 'user' in response.updated_user_object:
                        result['user_state_update'] = response.updated_user_object['user']
            if response.raw_response is not None:
                if isinstance(response.raw_response, YandexResponse):
                    result = response.raw_response.to_dict()
                else:
                    result['response'] = response.raw_response
                return result
            if response.voice is not None and response.voice != response.text:
                result['response']['tts'] = response.voice
            buttons = response.links or []
            for button in buttons:
                # avoid cyrillic characters in urls - they are not supported by Alice
                if 'url' in button:
                    button['url'] = tgalice.utils.text.encode_uri(button['url'])
            if response.suggests:
                buttons = buttons + [{'title': suggest} for suggest in response.suggests]
            for button in buttons:
                if not isinstance(button.get('hide'), bool):
                    button['hide'] = True
            result['response']['buttons'] = buttons
            if response.image_id:
                result['response']['card'] = {
                    'type': 'BigImage',
                    'image_id': response.image_id,
                    'description': response.text
                }
            elif response.image_url and self.image_manager:
                image_id = self.image_manager.get_image_id_by_url(response.image_url)
                if image_id:
                    result['response']['card'] = {
                        'type': 'BigImage',
                        'image_id': image_id,
                        'description': response.text
                    }
            if response.gallery is not None:
                result['response']['card'] = response.gallery.to_dict()
            if response.image is not None:
                result['response']['card'] = response.image.to_dict()
            return result
        elif source == SOURCES.FACEBOOK:
            if response.raw_response is not None:
                return response.raw_response
            result = {'text': response.text}
            if response.suggests or response.links:
                links = [{'type': 'web_url', 'title': link['title'], 'url': link['url']} for link in response.links]
                suggests = [{'type': 'postback', 'title': s, 'payload': s} for s in response.suggests]
                result = {
                    "attachment": {
                        "type": "template",
                        "payload": {
                            "template_type": "button",
                            "text": response.text,
                            "buttons": links + suggests
                        }
                    }
                }
            return result  # for bot.send_message(recipient_id, result)
        elif source == SOURCES.TEXT:
            result = response.text
            if response.voice is not None and response.voice != response.text:
                result = result + '\n[voice: {}]'.format(response.voice)
            if response.image_id:
                result = result + '\n[image: {}]'.format(response.image_id)
            if response.image_url:
                result = result + '\n[image: {}]'.format(response.image_url)
            if response.sound_url:
                result = result + '\n[sound: {}]'.format(response.sound_url)
            if len(response.links) > 0:
                result = result + '\n' + ', '.join(
                    ['[{}: {}]'.format(link['title'], link['url']) for link in response.links]
                )
            if len(response.suggests) > 0:
                result = result + '\n' + ', '.join(['[{}]'.format(s) for s in response.suggests])
            if len(response.commands) > 0:
                result = result + '\n' + ', '.join(['{{{}}}'.format(c) for c in response.commands])
            return [result, has_exit_command]
        elif source == SOURCES.VK:
            # todo: instead of a dict, use a class object as a response
            # todo: add multimedia, etc.
            result = {
                'text': response.text,
            }
            if response.suggests or response.links:
                rows = []
                for i, link in enumerate(response.links):
                    if i % self.tg_suggests_cols == 0:
                        rows.append([])
                    rows[-1].append({'action': {'type': 'open_link', 'label': link['title'], 'link': link['url']}})
                for i, suggest in enumerate(response.suggests):
                    if i % self.tg_suggests_cols == 0:
                        rows.append([])
                    rows[-1].append({'action': {'type': 'text', 'label': suggest}})
                for row in rows:
                    for button in row:
                        label = button['action']['label']
                        if len(label) > 40:
                            button['action']['label'] = label[:37] + '...'
                result['keyboard'] = {
                    'one_time': True,
                    'buttons': rows,
                }
            return result
        else:
            raise ValueError(SOURCES.unknown_source_error_message)

    def serverless_alice_handler(self, alice_request, context):
        """ This method can be set as a hanlder if the skill is deployed as a Yandex.Cloud Serverless Function """
        return self.respond(alice_request, source=SOURCES.ALICE)



from typing import Dict

from . import request, response
from .request import YandexRequest
from .response import YandexResponse


def extract_yandex_forms(req: YandexRequest) -> Dict[str, Dict[str, str]]:
    results = {}
    raw_forms = req and req.request and req.request.nlu and req.request.nlu.intents
    if not raw_forms:
        return results
    for intent_name, intent in raw_forms.items():
        results[intent_name] = {
            slot_name: slot.value
            for slot_name, slot in intent.slots.items()
        }
    return results



import distutils.util
import warnings

from html.parser import HTMLParser


class TTSParser(HTMLParser):
    """
    Allows parsing texts like
        'I study in the <text>1</text><voice>first</voice> grade. <a href="blood5.ru">The proof</a>'
    """
    TAG_TEXT = 'text'
    TAG_VOICE = 'voice'
    TAG_LINK = 'a'
    TAG_SPEAKER = 'speaker'  # it has no close tag
    SUPPORTED_TAGS = {TAG_TEXT, TAG_VOICE, TAG_LINK, TAG_SPEAKER}

    def __init__(self):
        super(TTSParser, self).__init__()
        self._text = ''
        self._voice = ''
        self._current_tag = None
        self._links = []

    def handle_starttag(self, tag, attrs):
        if self._current_tag is not None:
            raise ValueError('Open tag "{}" encountered, but tag "{}" is not closed'.format(self._current_tag, tag))
        if tag not in self.SUPPORTED_TAGS:
            warnings.warn('Encountered an unknown tag "{}", will ignore it'.format(tag))
        if tag == self.TAG_SPEAKER:
            self._voice += self.get_starttag_text()
            return  # speaker tag cannot be current
        self._current_tag = tag
        if tag == self.TAG_LINK:
            attrs_dict = dict(attrs)
            if 'href' not in attrs_dict:
                raise ValueError('The "a" tag has no "href" attribute; attrs: "{}".'.format(attrs))
            link = {'url': attrs_dict['href'], 'title': ''}
            if 'hide' in attrs_dict:
                link['hide'] = bool(distutils.util.strtobool(attrs_dict['hide']))
            self._links.append(link)

    def handle_endtag(self, tag):
        if self._current_tag is None:
            raise ValueError('Encountered close tag "{}", but there are no open tags'.format(tag))
        if tag == self.TAG_LINK:
            if self._links[-1]['title'].strip() == '':
                raise ValueError('The "a" tag has emtpy contents')
        self._current_tag = None

    def handle_data(self, data):
        if self._current_tag is None or self._current_tag == self.TAG_TEXT:
            self._text += data
        if self._current_tag is None or self._current_tag == self.TAG_VOICE:
            self._voice += data
        if self._current_tag == self.TAG_LINK:
            self._links[-1]['title'] += data

    def get_text(self):
        return self._text

    def get_voice(self):
        return self._voice

    def get_links(self):
        return self._links



import random


def make_unique(seq):
    """ Remove duplicates from a sequence (keep=first), without losing its ordering """
    seen = set()
    # todo: make it work with dicts as well
    seen_add = seen.add
    return [x for x in seq if not (x in seen or seen_add(x))]


def sample_at_most(seq, n=1):
    """ Sample min(n, len(seq)) unique elements from seq in random order """
    seq = list(set(seq))
    random.shuffle(seq)
    return seq[:n]



# -*- coding: utf-8 -*-
from __future__ import print_function

from tgalice import dialog, dialog_manager, interfaces, storage, nlu, nlg, testing, utils, dialog_connector
from tgalice.server import flask_server
from tgalice.storage import session_storage, message_logging
from tgalice.dialog_manager.base import COMMANDS
from tgalice.storage.message_logging import LoggedMessage
from tgalice.nlu import basic_nlu

from tgalice.dialog.names import COMMANDS, REQUEST_TYPES, SOURCES
